package NoRFC; # $Id: NoRFC.pm 183 2021-03-04 16:58:01Z minus $
use strict;
use utf8;
use feature qw/ say /;

=encoding utf8

=head1 NAME

NoRFC - Management library of noRFC documents

=head1 VERSION

Version 1.03

=head1 SYNOPSIS

    use NoRFC;

=head1 ABSTRACT

norfc - program which performs management of noRFC documents

=head1 DESCRIPTION

This module provides norfc functionality

=head2 again

Internal use only

=head2 extract

    my %extract_data = $app->extract(
            num => "RFC-R1001010",
            pub => 1614716051,
            category => "Etcetera",
            author => "Anonymous",
            title => "Test document",
            abstract => "Content",
            signature => "...xxx...",
            dir => "./",
        );

This method performs extraction the noRFC document and detached signature by number of specified NoRFC
Will be used local database directory as directory (dir) by default

Returns structure

    (
        docfile => "/full/path/to/document.txt",
        sigfile => "/full/path/to/document.txt.sig",
    )

 or error otherwise:

    die $app->error unless %extract_data;

=head2 gnupg

    my $gnupg = $app->gnupg();

Returns gnupg object

=head2 is_valid

    $app->is_valid("RFC-R1001010") or die ($self->error);

Returns 0 and sets error if specified document number is not valid,
and returns 1 if is valid

    $app->is_valid("RFC-R1001010", QUIET_MODE) or die ("Incorrect document number");

Returns 0 in quiet mode (without error setting) if specified document number is not valid,
and returns 1 if is valid

=head2 m11r_checkdigit_compute

    my $m11r = m11r_checkdigit_compute("123456789"); # 5
    my $m11r = $app->m11r_checkdigit_compute("123456789"); # 5
    my $m11r = __PACKAGE__->m11r_checkdigit_compute("123456789"); # 5

Returns M11R checkdigit by specified digits-string

See also L<Algorithm::CheckDigits::M11_015> and check_okpo()

=head2 parse_norfcnum

    my %norfcnum = $app->parse_norfcnum("Draft-R21050019");

Parses the noRFC document and returns as hash-structure:

    (
      'dir' => '21/21050',
      'm11r' => '9',
      'nn' => '01',
      'status' => 'R',
      'type' => 'Draft',
      'y3' => 21,
      'yday' => '050'
    )

=head2 private_keys_info

    my %users = $app->private_keys_info();

Returns structure that describes private keys in format:

    fingerprint => {name, validity, order}

=head2 raise

    $app->raise("Error string");

Prints error string to STDERR pipe

=head2 sign

    my $sign = $app->sign( "...fingerprint...", "Document body" )
        or die($app->error());

Signs document and returns detached signature

=head2 slurp_io

    my $data = $app->slurp_io(); # STDIO pipe
    my $data = $app->slurp_io($io); # IO::Handle and etc. pipes

Reads and returns all bytes from IO pipe. Default: STDIO pipe

=head2 store

    my $store = $app->store();

Returns store object

=head2 verify

    $app->verify($norfc_sig_file, $norfc_file) or die "Bad!"
    $app->verify($norfc_sig_file, $norfc_file, QUIET_MODE) or die "Bad!"

Verifies document and returns status. The QUIET_MODE flag suppress gpg pipes data

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 BUGS

* none noted

=head1 SEE ALSO

L<Algorithm::CheckDigits>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2021 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/$VERSION/;
$VERSION = '1.03';

use base qw/CTK::App/;

use Carp;
use Encode; # Encode::_utf8_on();
use Encode::Locale;

use Cwd;
use File::Spec;
use File::Path qw/rmtree/;
use Term::ANSIColor qw/colored/;
use YAML::XS qw/DumpFile LoadFile/;
use Try::Tiny;
use IO::Handle;
use GnuPG::Handles;
use GnuPG::Interface;

use CTK::Util qw/ trim dtf tz_diff variant_stf preparedir dformat bload bsave/;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;

use NoRFC::Store;

use constant {
        PROJECTNAME         => "NoRFC",
        PROJECTNAMEL        => "norfc",
        DATADIR             => "norfc",
        YFILE               => "norfc.yml",
        INDEXFILE           => "index.yml",
        DBFILE              => "norfc.db",
        DBDIR               => "db",
        SNAPSHOTDIR         => "snapshot",
        RFCNUM_BASE_FORMAT  => "%d%03d%02d",
        RFCNUM_FORMAT       => "%s-%s%d%03d%02d%d",
        RFCNUM_PATTERN      => qr/^(Draft|RFC)\-([RPXZ])([0-9]{1,3})([0-9]{3})([0-9]{2})([0-9]{1})$/,
        DATE_FORMAT         => "%YYYY-%MM-%DD",
        DATETIME_FORMAT     => "%YYYY-%MM-%DD %hh:%mm:%ss",
        DATETIME_GMT_FORMAT => "%YYYY-%MM-%DD %hh:%mm:%ss %G",
        DEFAULT_CATEGORY    => "Unknown",
        DEFAULT_TITLE       => "No title",
        DEFAULT_CONTENT     => "No content",
        DEFAULT_AUTHOR      => "Anonymous",
        QUIET_MODE          => 1,
        MAX_ROWS            => 1000, # Records per page
        IS_TTY              => (-t STDOUT) ? 1 : 0,
        LETS_MAP            => {
            R => "Regular",
            P => "Private",
            X => "Extra",
            Z => "Secret",
        },
        RFC_CATEGORIES      => [
            "Informational",
            "Experimental",
            "Standards Track",
            "Best Current Practice",
            "Historic",
            "Unknown",
        ],
        EXTRA_CATEGORIES    => [
            "Application",
            "Script",
            "Module",
            "WEB Site",
            "Server",
            "Service",
            "Configuration",
            "Etcetera",
        ],
    };

use constant NORFC_TEMPLATE => <<'TEMPLATE';
Document: [NORFC_RFCNUM]
Category: [NORFC_CATEGORY]
Published: [NORFC_PUB]
[NORFC_AUTHOR]

[NORFC_TITLE]

[NORFC_CONTENT]
TEMPLATE

use constant NORFC_USAGE => <<'USAGE';

Generation new noRFC document:
    norfc gen [<FILE>]
Show list of existing documents:
    norfc show
Show document information:
    norfc show <NORFC>
Check (validate) the document number:
    norfc check <NORFC>
Get the noRFC document:
    norfc get <NORFC>
Update document:
    norfc set <NORFC> [<FILE>]
Delete document:
    norfc del <NORFC>
Full documentation:
    norfc -H

USAGE

__PACKAGE__->register_handler(
    handler     => "version",
    description => sprintf("%s Version", PROJECTNAME),
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    printf("%s/%s\n", PROJECTNAME, $self->VERSION);
    return 1;
});

__PACKAGE__->register_handler(
    handler     => "info",
    description => "Get general information",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $store = $self->store;

    printf("Version         : %s\n", $self->VERSION);
    printf("Data dir        : %s\n", $self->datadir);
    printf("Status file     : %s\n", File::Spec->catfile($self->datadir, YFILE));
    printf("Config file     : %s\n", $self->conf("loadstatus") ? $self->configfile : _red("not loaded"));
    printf("DB DSN          : %s\n", $store->dsn);
    printf("DB status       : %s\n", $store->status ? _green("OK") : _red("ERROR"));
        $self->raise($store->error || sprintf("Store (%s): Unknown error", $store->dsn)) unless $store->status;

    # Fingerprint
    my $fingerprint = value($self->conf("fingerprint"));
    if ($fingerprint) {
        if (_fingerprint_test($fingerprint)) {
            printf("Fingerprint     : %s\n", $fingerprint);
        } else {
            $self->raise("The fingerprint is incorrect");
        }
    }

    # Total documents
    printf("Total documents : %s\n", $store->status ? $store->count : _yellow("unknown"));

    # Configuration
    if ($self->conf("loadstatus")) {
        $self->debug("Configuration   :");
        $self->debug(_explain($self->config));
    }

    print _cyan(NORFC_USAGE);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "gen",
    description => "Generation noRFC document",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    say "Generation new noRFC document";

    # Files
    my $yfile = File::Spec->catfile($self->datadir, YFILE);
    my $dbfile = File::Spec->catfile($self->datadir, DBFILE);

    # Document type
    my $type = "";
    if (my $t = $self->option("type")) {
        if ($t =~ /draft/i) {
            $type = "Draft";
        } elsif ($t =~ /rfc/i) {
            $type = "RFC";
        } else {
            $self->raise("Incorrect document type! Please specify value one from: Draft or RFC");
        }
    }
    unless ($type) {
        say "Available document types:";
        say "    R - RFC document";
        say "    D - Draft of RFC document (default)";
        my $t = trim($self->cli_prompt('Please provide the document type (R/D):','D'));
        if ($t =~ /^d/i) {
            $type = "Draft";
        } elsif ($t =~ /^r/i) {
            $type = "RFC";
        } else {
            $self->error("Incorrect document type! Please specify value one from: Draft or RFC");
            return 0;
        }
    }
    $self->debug(sprintf("Document type: %s", $type));

    # Letter
    my $let = "";
    $let .= "R" if $self->option("regular");
    $let .= "P" if $self->option("private");
    $let .= "X" if $self->option("extra");
    $let .= "Z" if $self->option("secret");
    unless ($let) {
        say "Available document statuses:";
        say "    R - Regular public document (default)";
        say "    P - Private document";
        say "    X - Extra";
        say "    Z - Top secret document";
        $let = uc(trim($self->cli_prompt('Please type a char of the status (R/P/X/Z):','R')));
    }
    if ($let !~ /^[RPXZ]$/) {
        $self->error("Incorrect the letter char selected! Please specify one flag -R, -P, -X or -Z only! Default: -R");
        return 0;
    }
    $self->debug(sprintf("Status letter: %s", $let));

    # Date data
    my ($mday, $mon, $year, $yday) = (localtime(time))[3,4,5,7];
    $mon++; $yday++; $year += 1900;

    # Year
    if (my $y = $self->option("year")) {
        if ($y =~ /^[0-9]{4}$/) {
            $year = $y;
        } else {
            $self->error("Incorrect year! Please specify real year");
            return 0;
        }
    }
    my $y3 = $year % 1000;
    $self->debug(sprintf("Year (Y3): %d (%d)", $year, $y3));
    $self->debug(sprintf("Month: %d", $mon));

    # Year's day
    if (my $d = $self->option("day")) {
        if (is_int($d) && $d && $d <= 366) {
            $yday = $d;
        } else {
            $self->error("Incorrect year's day! Please specify real year's day (value between 1 and 365 or 366)");
            return 0;
        }
    }
    $self->debug(sprintf("Day (YDAY): %d (%d)", $mday, $yday));

    # Key
    my $dkey = sprintf("%d-%02d-%02d", $year, $mon, $mday);
    $self->debug(sprintf("Key: %s", $dkey));

    # Sequence number (counter)
    my $cnt = 0;
    if (my $number = $self->option("number")) {
        if (is_int($number) && $number && $number <= 99) {
            $cnt = $number;
        } else {
            $self->error("Incorrect number value! Please specify value between 1 and 99");
            return 0;
        }
    } else {
        # Load history data
        my $history = {};
        if (-e $yfile and -s _) {
            try {
                $history = LoadFile($yfile);
            } catch {
                $self->error(sprintf("Can't load file %s: %s", $yfile, $_ // "unknown error"));
            };
            return 0 if $self->error;
        }
        $cnt = ($history->{$dkey} || 0) + 1;
        if ($cnt > 99) {
            $self->error("No free counter values. Try specify year or day manually");
            return 0;
        }
    }

    # Category name
    my $_category = $self->option("category");
    unless ($_category) {
        say "Available recommended document categories:";
        printf "    RFC cats: %s\n", join(", ", @{(RFC_CATEGORIES)});
        printf "    Extra cats: %s\n", join(", ", @{(EXTRA_CATEGORIES)});
        $_category = $self->cli_prompt('Please provide the category of the document:', DEFAULT_CATEGORY);
    }
    my $category = decode(locale => $_category);
    if (length($category) > 64) {
        $self->error("The category value is too long! Maximum: 64 chars");
        return 0;
    }
    $self->debug(sprintf("Category: %s", $_category));

    # Title
    my $_title = $self->option("title")
        || $self->cli_prompt('Please provide the title of the document:', DEFAULT_TITLE);
    my $title = decode(locale => $_title);
    if (length($title) > 255) {
        $self->error("The title value is too long! Maximum: 255 chars");
        return 0;
    }
    $self->debug(sprintf("Title: %s", $_title));

    # DB store
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Generate number
    my $result = "";
    while ($cnt <= 99) {
        $result = sprintf(RFCNUM_FORMAT, $type, $let, $y3, $yday, $cnt,
            m11r_checkdigit_compute(sprintf(RFCNUM_BASE_FORMAT, $y3, $yday, $cnt)));
        my %info = $store->get($result);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s) unknown error", $store->dsn) );
            return 0;
        }
        if ($self->option("force")) {
            $self->debug(sprintf("Counter (force): %d", $cnt));
            last;
        } elsif ($info{id}) { # Found
            $self->debug(sprintf("Counter (skipped): %d", $cnt));
        } else {
            $self->debug(sprintf("Counter: %d", $cnt));
            last;
        }
    } continue {
        $cnt++;
    }
    if ($cnt > 99) {
        $self->error("No free counter values. Try specify year or day manually");
        return 0;
    }
    $self->debug(sprintf("Document number: %s", $result));

    # Document file (abstract)
    my $file = shift(@arguments) // '';
    my $content = DEFAULT_CONTENT;
    if (!length($file)) {
        $self->debug("No file of document specified");
    } elsif (!(-e $file and -T $file)) {
        $self->error(sprintf("File not exists or incorrect: %s", $file));
        return 0;
    } else {
        $content = bload($file, 1);
    }
    Encode::_utf8_on($content);
    $self->debug(sprintf("Content length: %d", length($content)));

    # Fingerprint (config, as default)
    my $fp_default = value($self->conf("fingerprint"));
    if ($fp_default && !_fingerprint_test($fp_default)) {
        $self->error("The fingerprint is incorrect (see configuration file)");
        return 0;
    }
    $self->debug(sprintf("Default fingerprint: %s", $fp_default)) if $fp_default;

    # Fingerprint (from command line)
    my $fingerprint = $self->option("fingerprint");
    if ($fingerprint && !_fingerprint_test($fingerprint)) {
        $self->error("The user's fingerprint is incorrect (see command line syntax)");
        return 0;
    }
    $self->debug(sprintf("Fingerprint from command line: %s", $fingerprint)) if $fingerprint;

    # Get author
    my $author; # Author
    my %users = $self->private_keys_info();
    if ($fingerprint && exists($users{$fingerprint})) {
        $author = $users{$fingerprint}{name} || DEFAULT_AUTHOR;
        $self->debug(sprintf("@ %s", encode(locale => $author)));
    } elsif (%users) {
        say "Available secret-keys:\n";
        my $default_key;
        foreach my $k (sort {$users{$a}{order} <=> $users{$b}{order}} keys %users) {
            $fp_default ||= $k;
            $default_key = $k if $k eq $fp_default;
            printf("%s [%s]%s\n", _yellow($k), $users{$k}{validity}, $k eq $fp_default ? " (default)" : "");
            printf(" @ %s\n\n", encode(locale => $users{$k}{name}));
        }
        $fingerprint = $self->cli_prompt('Please select the secret-key (fingerprint):', $default_key);
        if ($fingerprint) {
            if ($fingerprint =~ /^\s*n/i) {
                $fingerprint = "";
                say _yellow("No signature specified. Skipped");
            } elsif (!(_fingerprint_test($fingerprint) && $users{$fingerprint})) {
                $self->error("Incorrect fingerprint");
                say _yellow("Type \"no\" to skip signing");
                return 0;
            }
        }
        $author = $users{$fingerprint}{name} || DEFAULT_AUTHOR;
        $self->debug(sprintf("@ %s", encode(locale => $author)));
    } else {
        $author = DEFAULT_AUTHOR;
    }
    $self->debug(sprintf("Fingerprint: %s", $fingerprint)) if $fingerprint;

    # Confirm
    my $pub = time();
    say "";
    say _magenta("Document number : %s", $result);
    say _magenta(" Type           : %s", $type);
    say _magenta(" Status         : %s", LETS_MAP()->{$let} // "");
    say _magenta(" Category       : %s", $_category);
    say _magenta(" Title          : %s", $_title);
    say _magenta(" Published      : %s", _fdate($pub));
    say _magenta(" Source file    : %s", $file) if $file;
    say _magenta(" Fingerprint    : %s", $fingerprint) if $fingerprint;
    if ($self->cli_prompt('All right?:','yes') !~ /^\s*y/i) {
        $self->error("Operation aborted");
        return 0;
    }

    # Create document
    my $doc = dformat(NORFC_TEMPLATE, {
        NORFC_RFCNUM    => $result,
        NORFC_CATEGORY  => $category,
        NORFC_PUB       => _fdate($pub, 1),
        NORFC_AUTHOR    => $author,
        NORFC_TITLE     => $title,
        NORFC_CONTENT   => $content,
    });
    Encode::_utf8_on($doc);
    my $norfc_file = File::Spec->catfile(getcwd(), sprintf("%s.txt", $result));
    if (bsave($norfc_file, $doc, 1)) {
        $self->debug(sprintf("File \"%s\" has been saved", $norfc_file));
    } else {
        $self->error(sprintf("Can't save file \"%s\"", $norfc_file));
        return 0;
    }

    # Sign
    my $sign = "";
    if ($fingerprint) {
        $sign = $self->sign($fingerprint, $doc) or return 0;
    }
    my $norfc_sig_file = File::Spec->catfile(getcwd(), sprintf("%s.txt.sig", $result));
    if (bsave($norfc_sig_file, $sign)) {
        $self->debug(sprintf("File \"%s\" has been saved", $norfc_sig_file));
    } else {
        $self->error(sprintf("Can't save file \"%s\"", $norfc_sig_file));
        return 0;
    }

    # Fix
    $store->add(
        num         => $result,
        pub         => $pub,
        category    => $category,
        title       => $title,
        author      => $author,
        abstract    => $content,
        signature   => $sign,
    ) or do {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0 unless $self->option("force");
    };


    # Save history data
    try {
        DumpFile($yfile, {$dkey => $cnt});
    } catch {
        $self->error(sprintf("Can't write file \"%s\": %s", $yfile, $_ // "unknown error"));
    };
    return 0 if $self->error && !$self->option("force");

    # Show result
    say "";
    say _cyan("The document %s has been successfully saved to\n %s", $result, $norfc_file);
    say _cyan("For verification, use:\n gpg --verify %s", $norfc_sig_file);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "get",
    description => "Extract noRFC document to current directory",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Get store instance
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Get number
    my $recipient = trim($self->option("recipient") || shift(@arguments)
        || $self->cli_prompt('Please provide the full document number (e.g., RFC-R1001010):'));
    return 0 unless $self->is_valid($recipient);

    # Get document info from DB
    my %info = $store->get($recipient);
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }
    unless ($info{id}) { # Found!
        $self->error("Document not found");
        return 0;
    }

    # Extract
    my %extract_data = $self->extract(%info, dir => getcwd());
    return 0 unless %extract_data;

    # Show result
    say "";
    say _cyan("The document %s has been successfully saved to\n %s", $recipient, $extract_data{docfile});
    say _cyan("For verification, use:\n gpg --verify %s", $extract_data{sigfile}) if -e $extract_data{sigfile};

    return 1;
});


__PACKAGE__->register_handler(
    handler     => "check",
    description => "Checking noRFC document",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $store = $self->store;

    # Get number
    my $recipient = trim($self->option("recipient") || shift(@arguments)
        || $self->cli_prompt('Please provide the full document number (e.g., RFC-R1001010):', "none"));
    $self->debug(sprintf("Document number : %s", $recipient));
    my $status = $self->is_valid($recipient, QUIET_MODE);

    # Get number digits
    my $digs;
    $digs = $1 if $recipient =~ /([0-9]+)/;
    unless ($digs && length($digs) > 1) {
        $self->error("The document number is incorrect");
        return 0;
    }

    # Parse number
    my ($test, $digit) = ($digs =~ /^(\d+?)(\d)$/);
    $self->debug(sprintf("Test string     : %s", $test));
    $self->debug(sprintf("Test status     : %s", $status ? "PASSED" : "FAILED"));
    $self->debug(sprintf("Got digit       : %d", $digit));
    my $m11r = m11r_checkdigit_compute($test);
    $self->debug(sprintf("Expected digit  : %d", $m11r));
    $status = 0 if $status && $digit != $m11r;

    # DB check
    if ($status) {
        my %info = $store->get($recipient);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
            return 0;
        }
        if ($info{id}) { # Found!
            $self->debug(sprintf("Found document  : %s", $info{num}));
            $self->debug(sprintf(" Id             : %d", $info{id}));
            $self->debug(sprintf(" Published      : %s", _fdate($info{pub})));
            $self->debug(sprintf(" Category       : %s", encode(locale => $info{category} || DEFAULT_CATEGORY)));
            $self->debug(sprintf(" Title          : %s", encode(locale => $info{title} || DEFAULT_TITLE)));
            $self->debug(sprintf(" Author         : %s", encode(locale => $info{author} || DEFAULT_AUTHOR)));

            # Extract
            my %extract_data = $self->extract(%info);
            if (%extract_data) {
                $self->debug(sprintf(" Document file  : %s", $extract_data{docfile}));
            } else {
                $self->raise("Can't extract noRFC document %s: %s", $recipient, $self->error || "Unknown error");
                $status = 0;
            }

            # Verify
            if (-e $extract_data{sigfile}) {
                $self->debug(sprintf(" Signature file : %s", $extract_data{sigfile}));
                unless ($self->verify($extract_data{sigfile}, $extract_data{docfile}, !$self->verbosemode)) {
                    $self->raise("Can't verify file \"%s\"", $extract_data{docfile});
                    $status = 0;
                }
            } else {
                say _magenta("No signature file \"%s\"", $extract_data{sigfile});
            }
        }  else {
            $self->raise("Document not found");
            return 0;
        }
    }

    # Summary status
    $self->debug(sprintf(" Summary status : %s", $status ? "PASSED" : "FAILED"));
    if ($status) {
        say _green("PASSED! Document is valid");
    } else {
        say _red("FAILED! Document is NOT valid");
        return 0;
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "show",
    description => "Search noRFC document by number or show list of documents",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # DB store
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Check document number and show information
    my $recipient = trim($self->option("recipient") || shift(@arguments) || "all");
    if ($recipient ne "all") { # By number
        $self->debug(sprintf("Information for %s", $recipient));
        return 0 unless $self->is_valid($recipient);

        # Get data from DB
        my %info = $store->get($recipient);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
            return 0;
        }
        printf("Document number : %s\n", $recipient);
        unless ($info{id}) {
            $self->error( "Document not found" );
            return 0;
        }

        # Show details
        printf(" Id             : %s\n", $info{id} || 0);
        printf(" Published      : %s\n", _fdate($info{pub}));
        printf(" Category       : %s\n", encode(locale => $info{category} || DEFAULT_CATEGORY));
        printf(" Title          : %s\n", encode(locale => $info{title} || DEFAULT_TITLE));
        printf(" Author         : %s\n", encode(locale => $info{author} || DEFAULT_AUTHOR));

        # Extract
        my %extract_data = $self->extract(%info);
        unless (%extract_data) {
            printf(" Verify         : %s\n", _yellow("SKIPPED"));
            return 0;
        }
        printf(" Document file  : %s\n", $extract_data{docfile});

        # Verify
        if (-e $extract_data{sigfile}) {
            printf(" Signature file : %s\n", $extract_data{sigfile});
            if ($self->verify($extract_data{sigfile}, $extract_data{docfile}, !$self->verbosemode)) {
                printf(" Verify         : %s\n", _green("PASSED"));
            } else {
                printf(" Verify         : %s\n", _red("FAILED"));
                $self->error(sprintf("Can't verify file \"%s\"", $extract_data{docfile}));
                return 0
            }
        } else {
            printf(" Verify         : %s\n", _yellow("SKIPPED"));
            say _magenta("No signature file \"%s\"", $extract_data{sigfile});
        }

        return 1;
    }

    # List!
    my $count = $store->count();
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }
    my $offset = $self->option("offset") || 0;
       $offset = 0 unless is_int($offset);
    my $row_count = $self->option("limit") || MAX_ROWS;
       $row_count = MAX_ROWS unless is_int($row_count);
    my $page = $self->option("page") || 0;
       $page = 0 unless is_int($page);
    my $pages = (($count - ($count % $row_count)) / $row_count) + (($count % $row_count) ? 1 : 0);

    # Test page
    if ($page && $page > $pages) {
        $self->error( sprintf("Incorrect page number specified. Range: 1-%d", $pages) );
        return 0;
    }
    if ($page && !$offset) {
        $offset = ($page - 1) * $row_count;
    } elsif (!$page && $offset) {
        $page = (( $offset - ($offset % $row_count)) / $row_count) + 1;
    }

    $self->debug(sprintf("Count (total)   : %d", $count));
    $self->debug(sprintf("Pages (total)   : %d", $pages));
    $self->debug(sprintf("Offset          : %s", $offset));
    $self->debug(sprintf("Row count       : %s", $row_count));
    $self->debug(sprintf("Page            : %d", $page || 1));

    # Get
    my @all = $store->getall($offset, $row_count);
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Show table
    my @arr;
    my $total = 0;
    my $totalf = "";
    print "------------------+---------------+--------------------------------------------\n";
format STDOUT_TOP =
 Document Number  | Category      | Title
------------------+---------------+--------------------------------------------
.
format STDOUT =
 @<<<<<<<<<<<<<<< | @<<<<<<<<<<<< | @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
@arr
.
format STDOUTBOT =
------------------+---------------+--------------------------------------------
@>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
$totalf
.
    foreach my $row (@all) {
        $total++;
        @arr = ();
        push @arr, $row->[1] || '';
        push @arr, variant_stf(encode(locale => $row->[3] || ''), 13);
        push @arr, variant_stf(encode(locale => $row->[4] || ''), 42);
        write;
    }
    $~ = "STDOUTBOT";
    $totalf = sprintf("row(s) %d of %d; page %d of %d", $total, $count, $page || 1, $pages || 1);
    write;

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "del",
    description => "Delete noRFC document by number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Check document number
    my $recipient = trim($self->option("recipient") || shift(@arguments)
        || $self->cli_prompt('Please provide the full document number (e.g., RFC-R1001010):', "none"));
    return 0 unless $self->is_valid($recipient);

    # Get store instance
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Delete
    $store->del($recipient) or do {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    };

    # Show result
    say _cyan("The document %s has been successfully removed", $recipient);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "truncate",
    description => "Delete all records from database",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Get store instance
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Request
    if ($self->cli_prompt(sprintf("Delete all records from database (%s)?:", $store->dsn), 'no') !~ /^\s*y/i) {
        $self->error("Operation aborted");
        return 0;
    }

    # Truncate
    $store->truncate() or do {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    };

    # Show result
    say _cyan("Database was cleared");

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "set",
    description => "Update the noRFC document by number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # DB store instance
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Get number
    my $recipient = trim($self->option("recipient") || shift(@arguments)
        || $self->cli_prompt('Please provide the full document number (e.g., RFC-R1001010):'));
    return 0 unless $self->is_valid($recipient);

    # Get document info from DB
    my %info = $store->get($recipient);
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }
    unless ($info{id}) { # Found!
        $self->error("Document not found");
        return 0;
    }

    # Category name
    my $_category = $self->option("category");
    unless ($_category) {
        say "Available recommended document categories:";
        printf "    RFC cats: %s\n", join(", ", @{(RFC_CATEGORIES)});
        printf "    Extra cats: %s\n", join(", ", @{(EXTRA_CATEGORIES)});
        $_category = $self->cli_prompt('Please provide the category of the document:', decode(locale => $info{category} || DEFAULT_CATEGORY));
    }
    my $category = decode(locale => $_category);
    if (length($category) > 64) {
        $self->error("The category value is too long! Maximum: 64 chars");
        return 0;
    }
    $info{category} = $category;

    # Title
    my $_title = $self->option("title")
        || $self->cli_prompt('Please provide the title of the document:', decode(locale => $info{title} || DEFAULT_TITLE));
    my $title = decode(locale => $_title);
    if (length($title) > 255) {
        $self->error("The title value is too long! Maximum: 255 chars");
        return 0;
    }
    $info{title} = $title;

    # Document file (abstract)
    my $file = pop(@arguments) // '';
    my $content = DEFAULT_CONTENT;
    if (!length($file)) {
        $self->debug("No file of document specified");
    } elsif (!(-e $file and -T $file)) {
        $self->error(sprintf("File not exists or incorrect: %s", $file));
        return 0;
    } else {
        $content = bload($file, 1);
    }
    Encode::_utf8_on($content);
    $info{abstract} = $content;

    # Fingerprint (config, as default)
    my $fp_default = value($self->conf("fingerprint"));
    if ($fp_default && !_fingerprint_test($fp_default)) {
        $self->error("The fingerprint is incorrect (see configuration file)");
        return 0;
    }

    # Fingerprint (from command line)
    my $fingerprint = $self->option("fingerprint");
    if ($fingerprint && !_fingerprint_test($fingerprint)) {
        $self->error("The user's fingerprint is incorrect (see command line syntax)");
        return 0;
    }

    # Get author
    my $author; # Author
    my %users = $self->private_keys_info();
    if ($fingerprint && exists($users{$fingerprint})) {
        $author = $users{$fingerprint}{name} || DEFAULT_AUTHOR;
    } elsif (%users) {
        say "Available secret-keys:\n";
        my $default_key;
        foreach my $k (sort {$users{$a}{order} <=> $users{$b}{order}} keys %users) {
            $fp_default ||= $k;
            $default_key = $k if $k eq $fp_default;
            printf("%s [%s]%s\n", _yellow($k), $users{$k}{validity}, $k eq $fp_default ? " (default)" : "");
            printf(" @ %s\n\n", encode(locale => $users{$k}{name}));
        }
        $fingerprint = $self->cli_prompt('Please select the secret-key (fingerprint):', $default_key);
        if ($fingerprint) {
            if ($fingerprint =~ /^\s*n/i) {
                $fingerprint = "";
                say _yellow("No signature specified. Skipped");
            } elsif (!(_fingerprint_test($fingerprint) && $users{$fingerprint})) {
                $self->error("Incorrect fingerprint");
                say _yellow("Type \"no\" to skip signing");
                return 0;
            }
        }
        $author = $users{$fingerprint}{name} || DEFAULT_AUTHOR;
    } else {
        $author = DEFAULT_AUTHOR;
    }
    $info{author} = $author;

    # Confirm
    my $pub = time();
    say "";
    say _magenta("Document number : %s", $recipient);
    say _magenta(" Id             : %s", $info{id});
    say _magenta(" Published      : %s", _fdate($pub));
    say _magenta(" Category       : %s", $_category);
    say _magenta(" Title          : %s", $_title);
    say _magenta(" Source file    : %s", $file) if $file;
    say _magenta(" Fingerprint    : %s", $fingerprint) if $fingerprint;
    if ($self->cli_prompt('All right?:','yes') !~ /^\s*y/i) {
        $self->error("Operation aborted");
        return 0;
    }

    # Create document
    my $doc = dformat(NORFC_TEMPLATE, {
        NORFC_RFCNUM    => $recipient,
        NORFC_CATEGORY  => $category,
        NORFC_PUB       => _fdate($pub, 1),
        NORFC_AUTHOR    => $author,
        NORFC_TITLE     => $title,
        NORFC_CONTENT   => $content,
    });
    Encode::_utf8_on($doc);
    my $norfc_file = File::Spec->catfile(getcwd(), sprintf("%s.txt", $recipient));
    unless (bsave($norfc_file, $doc, 1)) {
        $self->error(sprintf("Can't save file \"%s\"", $norfc_file));
        return 0;
    }
    $info{pub} = $pub;

    # Sign
    my $sign = "";
    my $norfc_sig_file = File::Spec->catfile(getcwd(), sprintf("%s.txt.sig", $recipient));
    if ($fingerprint) {
        $sign = $self->sign($fingerprint, $doc) or return 0;
        unless (bsave($norfc_sig_file, $sign)) {
            $self->error(sprintf("Can't save file \"%s\"", $norfc_sig_file));
            return 0;
        }
    }
    $info{signature} = $sign;

    # Set
    $info{num} ||= $recipient;
    my $st = $store->set(%info);
    unless ($st) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Show result
    say "";
    say _cyan("The document %s has been successfully saved to\n %s", $recipient, $norfc_file);
    say _cyan("For verification, use:\n gpg --verify %s", $norfc_sig_file) if -e $norfc_sig_file;

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "export",
    description => "Export all noRFC documents",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # DB store instance
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Get count
    my $count = $store->count();
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Get max_rows (limit)
    my $limit = $self->option("limit") || MAX_ROWS;
       $limit = MAX_ROWS unless is_int($limit);
    $self->debug(sprintf("User's max rows value: %d", $limit)) if $limit != MAX_ROWS;

    # Get pages number
    my $pages = (($count - ($count % $limit)) / $limit) + (($count % $limit) ? 1 : 0);
    my $pages_l = length($pages);

    # Load history data
    my $yfile = File::Spec->catfile($self->datadir, YFILE);
    my $dkey = dtf(DATE_FORMAT);
    my $history = {};
    if (-e $yfile and -s _) {
        try {
            $history = LoadFile($yfile);
        } catch {
            $self->error(sprintf("Can't load file %s: %s", $yfile, $_ // "unknown error"));
        };
        return 0 if $self->error;
    }

    # Make index data
    my %index_data = (
        dkey => $dkey, # Date from history
        dcnt => $history->{$dkey} || 0, # Day's counter from history
        rows => $count, # Number of documents in database
        max_rows => $limit, # Max rows perl page for importing
        pages => $pages, # Pages number
    );
    $self->debug(_explain(\%index_data));

    # Snapshot directory
    my $_user_dir = trim(shift(@arguments));
    my $snapshot_dir = $_user_dir || File::Spec->catdir($self->datadir, SNAPSHOTDIR);
    preparedir($snapshot_dir) or do {
        $self->error(sprintf("Can't prepare directory %s", $snapshot_dir));
        return 0;
    };
    rmtree($snapshot_dir, {keep_root => 1}) unless $_user_dir;

    # Save index data
    my $index_file = File::Spec->catfile($snapshot_dir, INDEXFILE);
    try {
        DumpFile($index_file, \%index_data);
    } catch {
        $self->error(sprintf("Can't write file \"%s\": %s", $index_file, $_ // "unknown error"));
    };
    return 0 if $self->error;

    # Get data
    my $step = $limit;
    my $page = 0;
    my $st = 1; # Ok by default
    say sprintf("Exporting %d documents (%d pages)", $count, $pages);
    for (my $offset = 0; $offset < $count; $offset += $step) {
        $page++;
        my $rows = ($offset + $step > $count) ? $count % $step : $step;
        #$self->debug( sprintf("> Page=%.*u; Offset=%05d; Rows=%04d", $pages_l, $page, $offset, $rows) );

        # Get data
        my @all = $store->getall($offset, $rows);
        unless ($store->status) {
            $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
            $st = 0;
        }

        # Save page file
        my $page_file = File::Spec->catfile($snapshot_dir, sprintf("%.*u.yml", $pages_l, $page));
        try {
            DumpFile($page_file, \@all);
        } catch {
            $self->error(sprintf("Can't write file \"%s\": %s", $page_file, $_ // "unknown error"));
            $st = 0;
        };

        say sprintf("Export page %.*u of %u... ", $pages_l, $page, $pages), $st ? _green("OK") : _red("ERROR");
        last unless $st;
    }

    # Ok
    say "\n", _cyan("All documents has been successfully exported to\n %s", $snapshot_dir) if $st;

    return $st;
});

__PACKAGE__->register_handler(
    handler     => "import",
    description => "Import all noRFC document",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # DB store instance
    my $store = $self->store;
    unless ($store->status) {
        $self->error( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
        return 0;
    }

    # Directories and files
    my $snapshot_dir = trim(shift(@arguments)) || File::Spec->catdir($self->datadir, SNAPSHOTDIR);
    my $yfile = File::Spec->catfile($self->datadir, YFILE);
    my $index_file = File::Spec->catfile($snapshot_dir, INDEXFILE);
    unless (-e $snapshot_dir) {
        $self->error( sprintf("Directory %s not exists", $snapshot_dir) );
        return 0;
    }

    # Load index data
    my $index_data = {};
    if (-e $index_file and -s _) {
        try {
            $index_data = LoadFile($index_file);
        } catch {
            $self->error(sprintf("Can't load file %s: %s", $index_file, $_ // "unknown error"));
        };
        return 0 if $self->error;
        $self->debug(_explain($index_data));
    } else {
        $self->error( sprintf("Index file %s not exists", $index_file) );
        return 0;
    }

    # Get pages number
    my $pages = $index_data->{pages} || 0;
    my $rows = $index_data->{rows} || 0;
    my $pages_l = length($pages);
    unless ($rows && $pages) {
        $self->error("No rows found. Nothing to do");
        return 0;
    }

    # Processing
    my $msk = "Import page %.*u of %u... ";
    my $st = 1; # Ok by default
    say sprintf("Importing %d documents (%d pages)", $rows, $pages);
    for (my $page = 1; $page <= $pages; $page++) {
        my $page_file = File::Spec->catfile($snapshot_dir, sprintf("%.*u.yml", $pages_l, $page));
        #$self->debug( sprintf("> Page=%.*u; Pages=%04d; File=%s", $pages_l, $page, $pages, $page_file) );
        my $data = [];


        # Load page data
        if (-e $page_file and -s _) {
            try {
                $data = LoadFile($page_file);
            } catch {
                $self->error(sprintf("Can't load file %s: %s", $page_file, $_ // "unknown error"));
                $st = 0;
            };
        } else {
            say sprintf($msk, $pages_l, $page, $pages), _yellow("SKIP");
            next;
        }
        $data ||= [];

        # Main loop
        foreach my $row (@$data) {
            my $num = $row->[1];
            next unless $num;

            # Get data from DB
            my %info = $store->get($num);
            unless ($store->status) {
                $self->raise( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
                next;
            }

            # Update or Added
            my %info_new = (
                num         => $num,
                pub         => $row->[2] || time(),
                category    => $row->[3] || DEFAULT_CATEGORY,
                title       => $row->[4] || DEFAULT_TITLE,
                author      => $row->[5] || DEFAULT_AUTHOR,
                abstract    => $row->[6] || DEFAULT_CONTENT,
                signature   => $row->[7] || '',
            );
            if ($info{id}) { # Update
                $store->set(%info_new) or do {
                    $self->raise( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
                    $st = 0;
                };
            } else { # Add
                $store->add(%info_new) or do {
                    $self->raise( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
                    $st = 0;
                };
            }
        }

        say sprintf($msk, $pages_l, $page, $pages), $st ? _green("OK") : _red("ERROR");
        last unless $st;
    }

    # Save history data
    my $dkey = $index_data->{dkey} || dtf(DATE_FORMAT);
    try {
        DumpFile($yfile, {$dkey => $index_data->{dcnt} || 0});
    } catch {
        $self->error(sprintf("Can't write file \"%s\": %s", $yfile, $_ // "unknown error"));
    };
    return 0 if $self->error;

    # Ok
    say "\n", _cyan("All documents has been successfully imported to\n %s", $store->dsn) if $st;

    return 1;
});


sub again {
    my $self = shift;
       $self->SUPER::again(); # CTK::App again first!!

    # Reset datadir from config file
    if (my $newdatadir = $self->config("datadir")) {
        $self->datadir($newdatadir);
    }

    # Prepare DataDir
    preparedir($self->datadir()) or do {
        $self->raise("Can't prepare directory %s", $self->datadir());
    };

    # Store
    my $store_conf = $self->config("store") || {};
       $store_conf = {} unless is_hash($store_conf);
    my %store_args = %$store_conf;
    $store_args{file} = File::Spec->catfile($self->datadir, DBFILE)
        unless ($store_args{file} || $store_args{dsn});
    my $store = NoRFC::Store->new(%store_args);
    $self->{store} = $store;
    #$self->debug(_explain($store));

    # Create GnuPG instance
    my $gnupg = GnuPG::Interface->new();
    $gnupg->options->hash_init(
        armor   => 1,
    );
    $self->{gnupg} = $gnupg;

    return $self; # CTK requires!
}
sub raise {
    my $self = shift;
    say STDERR _red(@_);
}
sub store {
    my $self = shift;
    return $self->{store}
}
sub gnupg {
    my $self = shift;
    return $self->{gnupg}
}
sub m11r_checkdigit_compute {
    # See also: Algorithm::CheckDigits::M11_015 and check_okpo()
    my $test = pop;
    croak "Incorrect input digit-string" if !$test || $test =~ m/[^0-9]/g;
    my $len = length($test);
    my $iters = ($len + (($len & 1) ? 1 : 0)) / 2;
    my @digits = split(//, $test); # Get all digits from input string of chars
    #printf "Test=%s; len=%d; iters=%d\n", $test, $len, $iters;

    my $w_lim = 10; # Maximum for round-robin(10) weight list: 1,2,3,4,5,6,7,8,9,10,1,2,3,4,5...
    my $step = 2; # Step for weight list offset calculation for next iteration

    # Calculation sum for one weight list by ofset
    my $calc = sub {
        my $off = shift || 0;
        my $s = 0;
        for (my $i = 0; $i < $len; $i++) {
            my $w = (($i + $off) % $w_lim) + 1;
            $s += ($w * $digits[$i]);
            #printf " > i=%d; d=%d; w=%d; sum=%d\n", $i, $digits[$i], $w, $s;
        }
        return $s % 11;
    };

    # Main cycle
    my $sum = 0;
    for (my $j = 0; $j < $iters; $j++) {
        my $offset = $j*$step;
        $sum = $calc->($offset);
        #printf " >> j=%d; offset=%d; sum=%d\n", $j, $offset, $sum;
        last if $sum < 10;
    }
    $sum = 0 if $sum >= 10; # 0 if incorrect again
    #printf " >>> sum=%d\n", $sum;

    return $sum;
}
sub parse_norfcnum {
    my $self = shift; # First argument
    my $number = shift; # Last argument
    my %ret = ();
    return %ret unless $number;

    if ($number =~ RFCNUM_PATTERN) {
        $ret{type} = $1;
        $ret{status} = $2;
        $ret{y3} = $3;
        $ret{yday} = $4;
        $ret{nn} = $5;
        $ret{m11r} = $6;
        $ret{dir} = File::Spec->catdir($ret{y3}, sprintf("%d%03d", $ret{y3}, $ret{yday})); # <Y3>/<Y3><YDAY>
    }

    return %ret;
}
sub is_valid {
    my $self = shift; # First argument
    my $number = shift; # Last argument
    my $verbose = shift ? 0 : 1; # 1 - silent; 0 - vrbose (default)
    unless ($number) {
        $self->error("No document number") if $verbose;
        return 0;
    }

    # Format valid
    my $test = 0;
    my $exp = 0;
    if ($number =~ RFCNUM_PATTERN) {
        unless ($3) { # Y3
            $self->error("The Y3 field of document number is incorrect") if $verbose;
            return 0;
        }
        unless ($4) { # YDAY
            $self->error("The YDAY field of document number is incorrect") if $verbose;
            return 0;
        }
        unless ($5) { # NN
            $self->error("The NN field of document number is incorrect") if $verbose;
            return 0;
        }
        $test = sprintf("%d%03d%02d", $3, $4, $5);
        $exp = $6; # Expected
    } else {
        $self->error("Document number format is NOT valid") if $verbose;
        return 0;
    }

    # M11R valid
    my $m11r = m11r_checkdigit_compute($test);
    if ($exp != $m11r) {
        $self->error("Document number is NOT valid") if $verbose;
        return 0;
    }

    return 1; # OK
}
sub slurp_io {
    my $self = shift;
    my $io = shift;
    unless ($io) {
        return "" if -t STDIN;
        $io = IO::Handle->new();
        return "" unless $io->fdopen(fileno(STDIN), "r");
    }

    # Slurp
    my $contents = join '', $io->getlines;
    $io->close;
    return $contents;
}
sub private_keys_info {
    my $self = shift;
    my %users = (); # fingerprint => {name, validity, order}
    my $order = 0;
    my @secret_keys = $self->gnupg->get_secret_keys();
    foreach my $key (@secret_keys) {
        my $fp = $key->fingerprint->as_hex_string;
        my $validity = sprintf("%s - %s", $key->creation_date_string || "Unknown", $key->expiration_date_string || "Never");
        my @user_ids = $key->user_ids;
        my @user_names = ();
        foreach my $user (@user_ids) {
            push @user_names, $user->as_string;
        }
        my $u_users = join("\n", @user_names);
        $users{$fp} = {
            name => decode(locale => $u_users),
            validity => $validity,
            order => $order++,
        };
    }
    return (%users);
}
sub extract {
    my $self = shift;
    my %args = @_;
    my %res = ();
    $self->error("");

    # num => "RFC-R1001010"
    my $norfcnum = $args{num};
    unless ($norfcnum) {
        $self->error("No document number");
        return ();
    }

    # Parse norfcnum
    my %d = $self->parse_norfcnum($norfcnum);

    # Create document
    my $doc = dformat(NORFC_TEMPLATE, {
        NORFC_RFCNUM    => $norfcnum,
        NORFC_PUB       => _fdate($args{pub}, 1),
        NORFC_CATEGORY  => $args{category} || DEFAULT_CATEGORY,
        NORFC_AUTHOR    => $args{author} || DEFAULT_AUTHOR,
        NORFC_TITLE     => $args{title} || DEFAULT_TITLE,
        NORFC_CONTENT   => $args{abstract} || DEFAULT_CONTENT,
    });
    Encode::_utf8_on($doc);

    # Directory
    my $norfc_dir = $args{dir} || File::Spec->catdir($self->datadir, DBDIR, $d{dir});
    preparedir($norfc_dir) or do {
        $self->error(sprintf("Can't prepare directory %s", $norfc_dir));
        return ();
    };

    # Filename
    my $norfc_file = File::Spec->catfile($norfc_dir, sprintf("%s.txt", $norfcnum));
    unless (bsave($norfc_file, $doc, 1)) {
        $self->error(sprintf("Can't save file \"%s\"", $norfc_file));
        return ();
    }

    # Signature
    my $norfc_sig_file = File::Spec->catfile($norfc_dir, sprintf("%s.txt.sig", $norfcnum));
    my $signature = $args{signature} // '';
    if (length($signature)) {
        unless (bsave($norfc_sig_file, $signature)) {
            $self->error(sprintf("Can't save file \"%s\"", $norfc_sig_file));
            return ();
        }
    }

    return (
        docfile => $norfc_file,
        sigfile => $norfc_sig_file,
    );
}
sub sign {
    my $self = shift;
    my $fp = shift;
    my $document = shift // '';
    $self->error("");
    unless (_fingerprint_test($fp)) {
        $self->error("Fingerprint incorrect");
        return;
    }
    unless (length($document)) {
        $self->error("No document");
        return;
    }

    my ($input, $output, $error) = (
            IO::Handle->new(),
            IO::Handle->new(),
            IO::Handle->new(),
        );
    my $handles = GnuPG::Handles->new(
            stdin    => $input,
            stdout   => $output,
            stderr   => $error,
        );
    $self->gnupg->options->extra_args([
            "--detach-sign",
            sprintf("--local-user=%s", $fp),
        ]);

    # This sets up the communication
    my $pid = $self->gnupg->sign( handles => $handles );

    # This passes in the plaintext
    #binmode($input, ':raw:utf8');
    $input->print(encode(locale => $document));
    $input->close; # this closes the communication channel, indicating we are done

    # Slurp oututs
    my $sign = $self->slurp_io($output);
    my $errs = $self->slurp_io($error);

    # Clean up the finished GnuPG process
    waitpid $pid, 0;

    # Check signature format
    if ($sign !~ /^\-\-\-\-\-BEGIN\sPGP.+\-\-\-\-\-END\sPGP/s) {
        $self->error($errs);
        return;
    }

    # Ok
    return $sign;
}
sub verify {
    my $self = shift;
    my $file_sig = shift;
    my $file_doc = shift;
    my $verbose = shift ? 0 : 1; # 1 - silent; 0 - vrbose (default)

    my ($input, $output, $error) = (
            IO::Handle->new(),
            IO::Handle->new(),
            IO::Handle->new(),
        );
    my $handles = GnuPG::Handles->new(
            stdin    => $input,
            stdout   => $output,
            stderr   => $error,
        );

    # This sets up the communication
    my $pid = $self->gnupg->wrap_call(handles => $handles, commands => ["--verify", $file_sig, $file_doc]);

    # Slurp oututs
    my $out = $self->slurp_io($output);
    say _yellow($out) if $verbose && $out;
    my $err = $self->slurp_io($error);

    # Clean up the finished GnuPG process
    waitpid $pid, 0;

    my $status = 0; # Error
    if ($err =~ /Good/) {
        say _yellow($err) if $verbose;
        $status = 1;
    } elsif ($err) {
        say _red($err) if $verbose;
    }
    if ($out =~ /Good/) {
        $status = 1;
    }

    return $status;
}

sub _fdate {
    my $d = shift || 0;
    my $g = shift || 0;
    return "unknown" unless $d;
    return dtf(DATETIME_GMT_FORMAT, $d, 1) if $g;
    return dtf(DATETIME_FORMAT . " " . tz_diff(), $d);
}
sub _fingerprint_test {
    my $fp = shift || '';
    my $l = length($fp);
    return 0 unless $l == 40 or $l == 16; # Fingerprint or KeyID
    return 1 if $fp =~ /^[0-9a-f]+$/i;
    return 0;
}

# Colored helper functions
sub _green {  IS_TTY ? colored(['bright_green'], sprintf(shift, @_)) : sprintf(shift, @_) }
sub _red {    IS_TTY ? colored(['bright_red'],    sprintf(shift, @_)) : sprintf(shift, @_) }
sub _yellow { IS_TTY ? colored(['bright_yellow'], sprintf(shift, @_)) : sprintf(shift, @_) }
sub _cyan {   IS_TTY ? colored(['bright_cyan'],   sprintf(shift, @_)) : sprintf(shift, @_) }
sub _blue {   IS_TTY ? colored(['bright_blue'],   sprintf(shift, @_)) : sprintf(shift, @_) }
sub _magenta {IS_TTY ? colored(['bright_magenta'],sprintf(shift, @_)) : sprintf(shift, @_) }

# Diagnostic functions
sub _explain {
    require Data::Dumper;
    my $dumper = Data::Dumper->new( [shift] );
    $dumper->Indent(1)->Terse(1);
    $dumper->Sortkeys(1) if $dumper->can("Sortkeys");
    return $dumper->Dump;
}

1;

__END__
